package net.wanji.com.netty;


import io.netty.bootstrap.ServerBootstrap;
import io.netty.channel.*;
import io.netty.channel.group.ChannelGroup;
import io.netty.channel.group.DefaultChannelGroup;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.nio.NioServerSocketChannel;
import io.netty.channel.socket.nio.NioSocketChannel;
import io.netty.handler.codec.LengthFieldBasedFrameDecoder;
import io.netty.util.concurrent.GlobalEventExecutor;
import lombok.extern.slf4j.Slf4j;
import net.wanji.com.cache.netty.NettyMessageCache;
import net.wanji.com.netty.handler.NettyServerHandler;
import net.wanji.com.pojo.netty.MessageResultPojo;
import org.apache.commons.lang3.StringUtils;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.scheduling.concurrent.ThreadPoolTaskExecutor;
import org.springframework.stereotype.Component;

import java.nio.ByteOrder;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

@Slf4j
@Component
public class NettyClient {

    private static final Map<String, ChannelId> IP_PORT_CHANNEL_ID_MAP = new HashMap<>();
    private static ChannelGroup channelGroup = new DefaultChannelGroup(GlobalEventExecutor.INSTANCE);

    /**
     * 建立连接
     */
    public static void start(int localPort, int remotePort, ThreadPoolTaskExecutor threadPoolExecutor) throws InterruptedException {
        final EventLoopGroup group = new NioEventLoopGroup();
        final EventLoopGroup child = new NioEventLoopGroup();
        ServerBootstrap bootstrap = new ServerBootstrap();
        bootstrap.group(group, child)
                .channel(NioServerSocketChannel.class)
                .childOption(ChannelOption.TCP_NODELAY, true)
                .childOption(ChannelOption.SO_KEEPALIVE, true)
                .handler(new ChannelInitializer<NioSocketChannel>() {
                    @Override
                    protected void initChannel(NioSocketChannel socketChannel) {
                        socketChannel.pipeline()
                                .addLast(new LengthFieldBasedFrameDecoder(ByteOrder.LITTLE_ENDIAN, Integer.MAX_VALUE, 4, 2, 0, 0, true))
                                .addLast(new NettyServerHandler(threadPoolExecutor));
                        channelGroup.add(socketChannel);
                        String key = StringUtils.join(socketChannel.remoteAddress().getHostString(), ":".intern(), socketChannel.remoteAddress().getPort());
                        log.info("client {} is connect success", key);
                        IP_PORT_CHANNEL_ID_MAP.put(key, socketChannel.id());
                    }
                });
        ChannelFuture channelFuture = bootstrap.bind(localPort).sync();
        log.info("server is started in {}", channelFuture.channel().localAddress());
        //收到停止信号
        Runtime.getRuntime().addShutdownHook(new Thread(() -> {
            //停掉所有的线程
            channelGroup.close().syncUninterruptibly();
            group.shutdownGracefully();
            child.shutdownGracefully();
            channelFuture.channel().closeFuture().syncUninterruptibly();
            log.warn("server is closed");
        }));
    }

    public static void sendMessage(String ip, int port, String message) {
        Channel channel = channelGroup.find(IP_PORT_CHANNEL_ID_MAP.get(StringUtils.join(ip, ":", port)));
        channel.writeAndFlush(message);
    }

    public static Object sendMessage(String ip, int port, String message, String command) {
        sendMessage(ip, port, message, command, 500);
        return null;
    }

    public static MessageResultPojo sendMessage(String ip, Integer port, String msg, String command, int timeout) {
        try {
            if (timeout > 0) {
                CountDownLatch countDownLatch = new CountDownLatch(1);
                String key = StringUtils.joinWith("#", ip, port, command);
                long now = LocalDateTime.now().toInstant(ZoneOffset.of("+8")).toEpochMilli();
                MessageResultPojo messageResultPojo = new MessageResultPojo(command, countDownLatch, now, timeout, null);
                NettyMessageCache.NETTY_MESSAGE_RESULT_MAP.put(key, messageResultPojo);
                sendMessage(ip, port, msg);
                countDownLatch.await(300, TimeUnit.MILLISECONDS);
                return NettyMessageCache.NETTY_MESSAGE_RESULT_MAP.remove(key);
            }
        } catch (InterruptedException e) {
            log.error("sendMsg is error", e);
        }
        return null;
    }

    @Scheduled(fixedRate = 300)
    public void checkWaitTimeout() {
        long now = LocalDateTime.now().toInstant(ZoneOffset.of("+8")).toEpochMilli();
        NettyMessageCache.NETTY_MESSAGE_RESULT_MAP.forEach((k, v) -> {
//            threadPoolExecutor.execute(() -> {
            if (now - v.getStartTime() > v.getWaitMillisecond()) {
                log.warn("wait {}ms {} timeout", v.getWaitMillisecond(), k);
                v.getCountDownLatch().countDown();
            }
        });
//        });
    }
}